# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tabulate import tabulate

class Message:
    """Text message view."""
    def show_data(self, text):
        """Print message.
        
        Parameters
        ----------
        text : str
            Text for printing.

        """
        print(text)
        

class Table:
    """Text table view."""
    def show_data(self, data, headers):
        """Show table with headers.

        Parameters
        ----------
        data: list of tuples
            Tabular data.
        headers: tuple
            Headers.
 
        """
        t = []
        for row in data:
            t_row = []
            for record in row:
                t_row.append(str(record)[:40])
            t.append(t_row)
        if headers:
            print('\n' + tabulate(t, headers=headers, tablefmt="orgtbl") + '\n')
        else:
            print('\n' + tabulate(t, tablefmt="orgtbl") + '\n')
       

class Exam:
    """Text view to show details of examination."""
    def show_data(self, exam):
        """Print information about examination.
        
        Parameters
        ----------
        exam: sme.Examination
            SME examination object.

        """
        s = '\nE: {} {} {} {}\n'.format(exam.name, exam.gender, exam.age, exam.diagnosis)
        for m in exam.ms:
            s += '    M: {}\n'.format(m.time)
        print(s)

class WhereExam:
    """Text view to show groups in which selected examination placed."""
    def show_data(self, group_records, headers, placed_in):
        """Show all groups and indicate where the object is located.

        Parameters
        ----------
        group_records : list of tuple
            Attributes of groups.
        headers : tuple
            Names of group attributes.
        placed_in : list of bool
            Indicators. True if examination placed in appropriate group, False overwise.

        """
        rows = [('X' if p else '', gr[0], gr[1]) for p, gr in zip(placed_in, group_records)]
        headers_ext = ('', ) + headers
        print('\n' + tabulate(rows, headers=headers_ext, tablefmt="orgtbl") + '\n')



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tkinter import *
from tkinter import ttk
from tkinter import messagebox

class SimpleSignal:
    """Provides simple signals and slots mechanism."""
    def __init__(self):
        """Constructor.

        Create empty list of connected function.
        """
        self.slots = []

    def emit(self, *args):
        """Emit signal."""
        for s in self.slots:
            if s:
                s(args)

    def connect(self, slot):
        """Connect signal with slot.
        
        Parameters
        ----------
        slot
            Function to be connected with signal.
        
        """
        for i in range(len(self.slots)):
            if self.slots[i] == slot:
                return
            if self.slots[i] == None:
                self.slots[i] = slot
                return
        self.slots.append(slot)

    def disconnect(self, slot):
        """Disconnect slot from signal.
        
        Parameters
        ----------
        slot
            Name of connected function.
        
        """
        for i in range(len(self.slots)):
            if self.slots[i] == slot:
                self.slots[i] = None


class TableWidget(Frame):
    """Table widget."""
    def __init__(self, parent, headers = []):
        """Constructor.

        Parameters
        ----------
        parent
            Master for widget.
        headers : list of str
            Headers for table.

        """
        Frame.__init__(self, parent)
        self.scrollbar = Scrollbar(self, orient=VERTICAL)
        self.tree = ttk.Treeview(self, yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.tree.yview)
        self.scrollbar.pack(side=RIGHT, fill=Y)
        self.tree.pack(side=LEFT, fill=BOTH, expand=True)
        self.set_columns(headers=headers)
        # signals
        self.item_opened = SimpleSignal()
        self.tree.bind("<Double-1>", self.item_opened.emit)
        self.tree.bind("<Return>", self.item_opened.emit)
        self.item_selected = SimpleSignal()
        self.tree.bind("<<TreeviewSelect>>", self.item_selected.emit)
        
    def selected_item_text(self):
        """Return text in selected item.

        Returns
        -------
        item_text : str.
            Text of selected item.
        
        """
        try:
            item = self.tree.selection()[0]
            item_text = self.tree.item(item,"text")
        except IndexError:
            item_text = None
        finally:
            return item_text

    def set_columns(self, **kwargs):
        """Different settings for columns.

        Parameters
        ----------
        headers : list of str
            Headers for table.
        anchors : list of str
            Anchors for columns. Use for align content.
        widths : list of int
            Widths of columns.

        """
        if 'headers' in kwargs:
            headers = kwargs['headers']
            self.column_ids = ["#0",] + headers[1:]
            self.tree["columns"] = headers[1:]
            for (cid, h) in zip(self.column_ids, headers):
                self.tree.heading(cid, text=h)
        if 'anchors' in kwargs:
            anchors = kwargs['anchors']
            for (cid, a) in zip(self.column_ids, anchors):
                self.tree.column(cid, anchor=a)
        if 'widths' in kwargs:
            widths = kwargs['widths']
            for (cid, w) in zip(self.column_ids, widths):
                self.tree.column(cid, width=w)
        self.tree.pack(side=LEFT, fill=BOTH, expand=True)

    def update_data(self, data):
        """Fill table with data.

        Parameters
        ----------
        data : list of tuple of str
            Data in table rows.
        
        """
        self.clear()
        for (i, row) in zip(range(len(data)), data):
            self.tree.insert("", END, iid = str(i), text=str(row[0]), values=row[1:])
        
    def clear(self):
        """Clear all items."""
        self.tree.delete(*self.tree.get_children())

    def remember_selection(self):
        """Remember selected item."""
        self._selected_item = self.tree.selection()[0]
        
    def restore_selection(self):
        """Restore selection."""
        if not self.tree.exists(self._selected_item):
            return
        self.tree.selection_set(self._selected_item)
        
class GroupingTable(TableWidget):
    """Table widget for grouping."""
    def __init__(self, parent):
        """Constructor.

        Parameters
        ----------
        parent
            Master for widget.
        """
        Frame.__init__(self, parent)
        headers = ["ID", "Name", "In"]
        widths = [50, None, 100]
        anchors=[None, None, CENTER]
        super().__init__(parent, headers)
        self.set_columns(widths=widths, anchors=anchors)
        self.tree.bind("<Double-1>", self.toggle_item)
        self.tree.bind("<Return>", self.toggle_item)

    def toggle_item(self, event):
        """Toggle item."""
        if len(self.tree.selection()) == 0:
            return
        item = self.tree.selection()[0]
        values = self.tree.item(item, 'values')
        self.tree.item(item, values=(values[0], 'X' if values[1] == '' else ''))
        
    def checked_group_ids(self):
        """Get IDs of checked groups.
        
        Returns
        -------
        groups_ids : list of str
            IDs of all groups.
        placed_in : list of bool
            True if exam in group, False otherwise.

        """
        placed_in = []
        group_ids = []
        for item in self.tree.get_children():
            group_ids.append(self.tree.item(item, 'text'))
            placed_in.append(self.tree.item(item, 'values')[1] == 'X')
        return group_ids, placed_in
        
class Message:
    """Text message view."""
    def show_data(self, text):
        """ Show message box.

        Parameters
        ----------
        text : str
            Message text.

        """
        messagebox.showinfo("Message", text)

class Storage(TableWidget):
    """Table widget for stoarge info."""
    def __init__(self, parent):
        """Constructor.

        Parameters
        ----------
        parent
            Master for widget.

        """
        headers = ["ID", "Name", "Description", "Num"]
        super().__init__(parent, headers)
        widths = [50, None, None, 50]
        anchors = [CENTER, None, None, CENTER]
        self.set_columns(widths=widths, anchors=anchors)
        self.last_group_id = None
        
    def show_data(self, data, headers):
        """Fill table with data.

        Parameters
        ----------
        data: list of tuples
            Tabular data.
        headers: tuple
            Headers.

        """
        self.update_data(data)

class Group(TableWidget):
    """Table widget for stoarge info."""
    def __init__(self, parent):
        """Constructor.
        
        Parameters
        ----------
        parent
            Master for widget.

        """
        headers = ["ID", "Name", "Diagnosis", "Age", "Gender"]
        super().__init__(parent, headers)
        widths = [50, None, None, None, None]
        anchors = [None, None, None, CENTER, CENTER]
        self.set_columns(widths=widths, anchors=anchors)
        
    def show_data(self, data, headers):
        """Fill table with data.

        Parameters
        ----------
        data: list of tuples
            Tabular data.
        headers: tuple
            Headers.

        """
        self.update_data(data)
        
class WhereExam(GroupingTable):
    """Tk view to show groups where examination is."""
    
    def show_data(self, group_records, headers, placed_in):
        """Show data.

        Parameters
        ----------
        group_records : list of tuple
            Attributes of groups.
        headers : tuple
            Names of group attributes.
        placed_in : list of bool
            Indicators. True if examination placed in appropriate group, False overwise.

        """
        rows = [
            (gr[0], gr[1], 'X' if p else '')
            for p, gr in zip(placed_in, group_records)
        ]
        headers_ext = headers + ('', )
        self.update_data(rows)



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Mapping examinations to and from JSON folders.

JSON folder is the folder that consists file info.json and signals in simple text format."""

import json
import numpy as np
import os
from collections import OrderedDict

from . import sme

def get_exam_from_folder(folder_name):
    """Get examination object from folder with file info.json.

    Parameters
    ----------
    folder_name : str
        Name of JSON folder.
    
    Return
    ------
    : sme.Examination
        Examination instance.

    """
    e = sme.Examination()
    with open('{}/info.json'.format(folder_name), 'r') as f:
        data = json.load(f)
            
    e.name = data['name']
    e.age = data['age']
    e.gender = data['gender']
    e.diagnosis = data['diagnosis']
    e.ms = []
    for m_data in data['measurements']:
        m = sme.Measurement()
        m.time = m_data['time']
        m.ss = []
        for s_data in m_data['signals']:
            s = sme.Signal()
            s.dt = s_data['dt']
            s.x = np.loadtxt('{}/{}'.format(folder_name, s_data['file']))
            m.ss.append(s)
        e.ms.append(m)
    return e

def put_exam_to_folder(e, folder_name):
    """Put examination to folder. Return True if sucsess, False overwise.

    Create info.json and text files with samples.

    Parameters
    ----------
    e : sme.Examination
        Examination instance.
    folder_name : str
        Folder name.
    
    """
    abs_folder_path = os.path.expanduser(folder_name)
    os.makedirs(abs_folder_path)
    e_dict = OrderedDict()
    e_dict['name'] = e.name
    e_dict['diagnosis'] = e.diagnosis
    e_dict['age'] = e.age
    e_dict['gender'] = e.gender
    e_dict['measurements'] = []
    for (m, mn) in zip(e.ms, range(1, len(e.ms)+1)):
        m_dict = OrderedDict()
        m_dict['time'] = m.time
        m_dict['signals'] = []
        for (s, sn) in zip(m.ss, range(1, len(m.ss)+1)):
            s_dict = OrderedDict()
            s_dict['dt'] = s.dt
            s_dict['file'] = 'signal-{}{}.txt'.format(mn, sn)
            np.savetxt("{}/signal-{}{}.txt".format(abs_folder_path, mn, sn), s.x, fmt='%f')            
            m_dict['signals'].append(s_dict)
        e_dict['measurements'].append(m_dict)
                           
    s = json.dumps(e_dict, ensure_ascii=False, indent='    ')
    with open('{}/info.json'.format(abs_folder_path), 'w') as f:
        f.write(s)



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sqlite3
import numpy as np
from collections import OrderedDict
import os

from .base_model import BaseModel
from . import sme
from . import sme_json_folders as jsme

class Model(BaseModel):
    """Model implementation for SQLite3 SME database."""
    
    def __init__(self):
        """Constructor.
        
        Create fields and set initial state of model.
        
        """
        super().__init__()
        self.conn = None
        self.c = None
        self.set_state(storage_opened=False, file_name = None)

    def create_storage(self, file_name):
        """Create new database.

        Parameters
        ----------
        file_name : str
            File name.

        """
        if self.state()['storage_opened']:
            self.close_storage()
        if os.path.isfile(file_name):
            os.remove(file_name)
        self.conn = sqlite3.connect(file_name)
        self.c = self.conn.cursor()
        self.c.executescript(create_sme_db_script)
        self.conn.commit()
        self.set_state(storage_opened=True, file_name = file_name)

    def open_storage(self, file_name):
        """Open database.

        Parameters
        ----------
        file_name : str
            File name.

        """
        if self.state()['storage_opened']:
            self.close_storage()
        self.conn = sqlite3.connect(file_name)
        self.c = self.conn.cursor()
        self.c.execute("pragma foreign_keys=on")
        self.set_state(storage_opened=True, file_name=file_name)

    def close_storage(self):
        """Close current database."""
        if self.state()['storage_opened']:
            self.conn.close()
            self.set_state(storage_opened=False, file_name=None)

    def storage_exists(self, file_name):
        """Check if the file exists.

        Parameters
        ----------
        file_name : str
            Name of database file.

        Returns
        -------
        : bool
            True if exists, False otherwise.

        """
        return os.path.isfile(file_name)

    @BaseModel.do_if_storage_opened
    def exam(self, exam_id):
        """Return examination from database.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        Returns
        -------
        sme.Examination
            Examination object.

        """
        e = sme.Examination()
        self.c.execute("""
        SELECT E.name, E.diagnosis, E.age, E.gender FROM examination AS E
        WHERE exam_id = ? """, [exam_id, ])
        result = self.c.fetchone()
        if not result:
            return None
        e.name, e.diagnosis, e.age, e.gender = result
        e.ms = []
        self.c.execute("""
        SELECT M.meas_id, M.time FROM measurement AS M
        WHERE exam_id = ?
        ORDER BY meas_id """, [exam_id, ])
        for m_sql in self.c.fetchall():
            m = sme.Measurement()
            m_id, m.time = m_sql
            m.ss = []
            self.c.execute("""
            SELECT S.dt, S.data FROM signal AS S
            WHERE meas_id = ? """, [m_id, ])
            for s_sql in self.c.fetchall():
                s = sme.Signal()
                s.dt = s_sql[0]
                s.x = np.array(np.frombuffer(s_sql[1]))
                m.ss.append(s)
            e.ms.append(m)
        return e

    @BaseModel.do_if_storage_opened
    def insert_exam(self, exam):
        """Insert examination into current database.

        Parameters
        ----------
        exam : sme.Examination
            Examination object

        """
        self.c.execute("""
        SELECT max(exam_id)
        FROM examination """)
        exam_id = self.c.fetchone()[0]
        if not exam_id:
            exam_id = 0

        self.c.execute("""
        SELECT max(meas_id)
        FROM measurement """)
        meas_id = self.c.fetchone()[0]
        if not meas_id:
            meas_id = 0

        self.c.execute("""
        INSERT INTO examination (name, diagnosis, age, gender)
        VALUES (?,?,?,?) """, (exam.name, exam.diagnosis, exam.age, exam.gender) )
        exam_id += 1

        for m in exam.ms:
            self.c.execute("""
            INSERT INTO measurement (time, exam_id)
            VALUES (?,?) """, (m.time, exam_id) )
            meas_id += 1

            for s in m.ss:
                self.c.execute("""
                INSERT INTO signal (data, dt, meas_id)
                VALUES (?,?,?) """, (s.x.tobytes(), s.dt, meas_id) )
            
        self.conn.commit()

    @BaseModel.do_if_storage_opened
    def delete_exam(self, exam_id):
        """Remove examination from database.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        """
        self.c.execute("""
        DELETE FROM Examination
        WHERE exam_id = ?
        """,(exam_id, ))
        self.conn.commit()

    @BaseModel.do_if_storage_opened
    def storage_info(self):
        """Return tabular information about database.

        Returns
        -------
        ext_data : list of tuple
            Table with information about database.
        ext_headers : tuple
            Headers.

        """
        self.c.execute("""
        SELECT * FROM egeg_group;
        """)
        data = self.c.fetchall()
        headers = tuple(map(lambda x: x[0], self.c.description))
        num_in_groups = []

        ext_data = []
        for d in data:
            self.c.execute("""
            SELECT COUNT(E.exam_id)
            FROM examination as E, group_element as GE
            WHERE GE.exam_id = E.exam_id AND GE.group_id = ? """, [d[0], ])
            n = self.c.fetchall()[0][0]
            ext_data.append(d + (n,))
        ext_headers = headers + ("number",)

        self.c.execute("""
        SELECT COUNT(exam_id)
        FROM examination
        WHERE exam_id NOT IN (SELECT exam_id FROM group_element) """)
        ungrouped_num = self.c.fetchall()[0][0]
        
        last_string = ['' for h in ext_headers]
        last_string[0] = '0'
        last_string[-2] = 'Ungrouped'
        last_string[-1] = ungrouped_num
        ext_data.append(tuple(last_string))
        
        return ext_data, ext_headers    

    @BaseModel.do_if_storage_opened
    def group_info(self, group_id):
        """Return short information about examinations in selected group.

        Parameters
        ----------
        group_id : str
            Group ID

        Returns
        -------
        data : list of tuple
            Examination descriptions.
        headers : tuple
            Headers.

        """
        if group_id == '0':
            self.c.execute("""
            SELECT exam_id, name, diagnosis, age, gender
            FROM examination
            WHERE exam_id NOT IN (SELECT exam_id FROM group_element) 
            """)
        else:
            self.c.execute("""
            SELECT COUNT(*) FROM egeg_group WHERE group_id = ?
            """, [group_id])
            if self.c.fetchall()[0][0] == 0:
                return None, None

            self.c.execute("""
            SELECT E.exam_id, E.name, E.diagnosis, E.age, E.gender
            FROM examination AS E, group_element AS GE
            WHERE GE.exam_id = E.exam_id AND GE.group_id = ?
            """, [group_id, ])
            
        data = self.c.fetchall()
        headers = tuple(map(lambda x: x[0], self.c.description))
        return data, headers

    @BaseModel.do_if_storage_opened
    def insert_group(self, name, description):
        """Add new group of examinations.

        Parameters
        ----------
        name : str
            Name of new group.
        description : str
            Description for new group.

        """
        self.c.execute("""
        INSERT INTO egeg_group (name, description)
        VALUES (?, ?) """, [name, description, ])
        self.conn.commit()

    @BaseModel.do_if_storage_opened
    def delete_group(self, group_id):
        """Delete group of examinations from database.

        Parameters
        ----------
        group_id : str
            Group ID.

        """
        self.c.execute("""
        DELETE FROM egeg_group
        WHERE group_id = ? """, [group_id, ])
        self.conn.commit()

    @BaseModel.do_if_storage_opened
    def group_exam(self, exam_id, group_ids, placed_in):
        """Add and delete examination to and from groups.

        Parameters
        ----------
        exam_id : str
            Examination ID.
        group_ids : list of str
            Group IDs.
        placed_in : list of bool
            True for examinations to be placed in groups. Length of group_ids must be equal to length of placed_in.

        """
        for (group_id, p) in zip(group_ids, placed_in):
            if p:
                try:
                    self.c.execute("""
                    INSERT OR IGNORE INTO group_element
                    VALUES (?, ?) """, [exam_id, group_id, ])
                except sqlite3.IntegrityError:
                    pass
            else:
                try:
                    self.c.execute("""
                    DELETE FROM group_element
                    WHERE exam_id = ? AND group_id = ? """, [exam_id, group_id, ])
                except sqlite3.IntegrityError:
                    pass
        self.conn.commit()

    @BaseModel.do_if_storage_opened
    def where_exam(self, exam_id):
        """Return description of groups where examination in or not in.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        Returns
        -------
        group_records : list of tuple
            All group records.
        headers : list of str
            Names of group attributes.
        placed_in : list of bool
            True if exam in group, False otherwise.
        
        """

        including_exam_groups_ids = [
            r[0] for r in self.c.execute("""
            SELECT G.group_id
            FROM egeg_group as G, group_element
            WHERE G.group_id = group_element.group_id AND group_element.exam_id = ? """, [exam_id])
        ]

        self.c.execute("SELECT * FROM egeg_group")
        group_records = self.c.fetchall()
        headers = tuple(map(lambda x: x[0], self.c.description))
        placed_in = [gr[0] in including_exam_groups_ids for gr in group_records]
        
        return group_records, headers, placed_in
    
    @BaseModel.do_if_storage_opened
    def add_sme_db(self, file_name):
        """Add SME sqlite3 database to current database.

        Parameters
        ----------
        file_name : str
            Name of file with SQLite3 SME database.

        """
        source_name = file_name
        dest_name = self.state()['file_name']
        if self.state()['storage_opened']:
            self.close_storage()
        SMEDBImporter().DBimport(dest_name, source_name)
        self.open_storage(dest_name)

    @BaseModel.do_if_storage_opened
    def add_gs_db(self, file_name):
        """Add GS SQLite3 database to current database.
        
        Parameters
        ----------
        file_name : str
            Name of file with SQLite3 Gastroscan database.

        """
        source_name = file_name
        dest_name = self.state()['file_name']
        if self.state()['storage_opened']:
            self.close_storage()
        GSDBImporter().DBimport(dest_name, source_name)
        self.open_storage(dest_name)

    @BaseModel.do_if_storage_opened
    def add_exam_from_json_folder(self, folder_name):
        """Add examination from JSON folder to current database.

        Parameters
        ----------
        folder_name : str
            Name of folder with info.json and signal txt files.
        
        """
        e = jsme.get_exam_from_folder(folder_name)
        self.insert_exam(e)

    @BaseModel.do_if_storage_opened
    def export_as_json_folder(self, exam_id, folder_name):
        """Export examination to JSON folder. Return True if sucsess, False otherwise.
        
        Parameters
        ----------
        exam_id: str
            Examination ID.
        folder_name : str
            Name of folder for export info.json and signals in txt format.
        
        """
        e = self.exam(exam_id)
        jsme.put_exam_to_folder(e, folder_name)

    @BaseModel.do_if_storage_opened
    def group_record(self, group_id):
        """Return attribute names and values of selected group.

        Parameters
        ----------
        group_id : str
            Group ID.

        Returns
        -------
        : OrderedDict
            Attributes names and values for selected group.

        """
        attr = OrderedDict()
        self.c.execute("""
        SELECT name, description
        FROM egeg_group
        WHERE group_id = ?
        """, (group_id, ))
        attr['name'], attr['description'] = self.c.fetchone()
        return attr

    @BaseModel.do_if_storage_opened
    def update_group_record(self, group_id, attr):
        """Update group record in database.

        Parameters
        ----------
        group_id : str
            Group ID.
        attr : OrderedDict
            Attributes names and values.

        """
        self.c.execute("""
        UPDATE egeg_group
        SET name = ?, description = ?
        WHERE group_id = ?
        """, (attr['name'], attr['description'], group_id, ))
        self.conn.commit()

class DBImporter:
    def DBimport(self, dest_filename, source_filename):
        """
        Import data.
        """	
        self._source_filename = source_filename
        self._dest_filename = dest_filename
        self.run()

class SMEDBImporter(DBImporter):
    def run(self):
        """Import."""
        self._dconn = sqlite3.connect(self._dest_filename)
        self._sconn = sqlite3.connect(self._source_filename)
        
        self._dest_c = self._dconn.cursor()
        self._src_c = self._sconn.cursor()
        self._dest_c.execute("attach database ? as 'source';", (self._source_filename,))
        self._dest_c.executescript(add_sme_db_script)
        self._dest_c.execute("detach database source;")
        self._dest_c.execute('drop table variable;')
        self._dconn.commit()
        
        self._sconn.close()
        self._dconn.close()

class GSDBImporter(DBImporter):
    def run(self):
        """Import."""
        self._dconn = sqlite3.connect(self._dest_filename)
        self._sconn = sqlite3.connect(self._source_filename)
        
        self._dest_c = self._dconn.cursor()
        self._src_c = self._sconn.cursor()
		
        res = self._src_c.execute('\
        SELECT name, diagnosis, age, sex, date, signal, edited \
        FROM record as r, waveform as w\
        WHERE r.id = w.record_id\
        ')
        
        exam_id, meas_id = self._get_last_ids()
        ind = []
        t = []
        for r in res:
            if self._form_exam(r, t, ind):
                exam_id, meas_id = self._insert_exam(t, ind, exam_id, meas_id)
                t = []
                ind = []
                t.append(r)
                ind.append(1)
        self._insert_exam(t, ind, exam_id, meas_id)
        self._dconn.commit()
        
        self._sconn.close()
        self._dconn.close()
        
    def _get_last_ids(self):
        """Get during (last) measurement's and examination's id."""
        self._dest_c.execute('SELECT max(exam_id) FROM examination')
        exam_id = self._dest_c.fetchone()[0]
        self._dest_c.execute('SELECT max(meas_id) FROM measurement')
        meas_id = self._dest_c.fetchone()[0]
        if not exam_id:
            exam_id = 0
        if not meas_id:
            meas_id = 0
        return (exam_id, meas_id)

    def _insert_exam(self, t, ind, exam_id, meas_id):
        """Insert examination to destination DB. Data takes from temporary lists t and ind."""
        dt = 0.5
        self._dest_c.execute('\
        INSERT INTO examination (name, diagnosis, age, gender)\
        VALUES (?,?,?,?)',(t[0][0],t[0][1],t[0][2],t[0][3]))
        exam_id += 1
        meas_id += 1
        self._dest_c.execute('\
        INSERT INTO measurement(time, exam_id)\
        VALUES (?,?)',(t[0][4],exam_id))
        self._dest_c.execute('\
        INSERT INTO signal(data, dt, edited,meas_id)\
        VALUES (?,?,?,?)',(t[0][5], dt, t[0][6],meas_id)) 
        if 2 in ind:
            self._dest_c.execute('\
            INSERT INTO signal(data, dt, edited,meas_id)\
            VALUES (?,?,?,?)',(t[ind.index(2)][5], dt, t[ind.index(2)][6],meas_id)) 
        if 3 in ind:
            self._dest_c.execute('\
            INSERT INTO measurement(time, exam_id)\
            VALUES (?,?)',(t[ind.index(3)][4],exam_id))
            meas_id += 1
            self._dest_c.execute('\
            INSERT INTO signal(data, dt, edited, meas_id)\
            VALUES (?,?,?,?)',(t[ind.index(3)][5], dt, t[ind.index(3)][6],meas_id)) 
        if 4 in ind:
            self._dest_c.execute('\
            INSERT INTO signal(data, dt, edited, meas_id)\
            VALUES (?,?,?,?)',(t[ind.index(4)][5], dt, t[ind.index(4)][6],meas_id)) 
        return (exam_id, meas_id)

    def _form_exam(self, r, t, ind):
        """Form temporary lists t and ind."""
        if len(t) == 0:
            t.append(r)
            ind.append(1)
        else:
            if t[-1][0] == r[0] and t[-1][2] == r[2] and t[-1][3] == r[3] and t[-1][4][:9] == r[4][:9]:
                if ind[-1] == 1 and t[-1][4] == r[4]:
                    ind.append(2)
                elif ind[-1] == 1 or ind[-1] == 2:
                    ind.append(3)
                elif ind[-1] == 3:
                    ind.append(4)
                t.append(r)
                return False
            else:
                return True


#---------------------------------------
# sqlite3 scripts
#---------------------------------------

# create new SME database 
create_sme_db_script = """
pragma foreign_keys=1;

create table signal(
	signal_id integer,
	data blob,
	dt real,
	edited integer, 
	meas_id integer references measurement(meas_id) on delete cascade, 
	primary key(signal_id)
);

create table measurement(
	meas_id integer, 
	time text, 
	exam_id integer references examination(exam_id) on delete cascade, 
	primary key(meas_id)
);

create table examination(
	exam_id integer, 
	name text,
	diagnosis text, 
	age integer, 
	gender text, 
	primary key(exam_id)
);

create table egeg_group(
	group_id integer, 
	name text, 
	description text,
	primary key(group_id)		
);

create table group_element(
	exam_id integer references examination(exam_id) on delete cascade, 
	group_id integer references egeg_group(group_id) on delete cascade, 
	primary key(exam_id, group_id)
);
"""

# import SME database
add_sme_db_script = """
-- Create temporary table for variables and store max values of SMEP entities from nation db.
drop table if exists variable;
create table variable(name text primary key, value integer);
insert into variable(name, value) values('max_exam_id', (select coalesce(max(exam_id),0) from examination));
insert into variable(name, value) values('max_meas_id', (select coalesce(max(meas_id),0) from measurement));
insert into variable(name, value) values('max_group_id', (select coalesce(max(group_id),0) from egeg_group));

-- Paste groups with increased id to onation DB from source DB
insert into egeg_group(group_id, name, description) select group_id + (select value from variable where name = 'max_group_id'), name, description from source.egeg_group;

-- Paste examinations
insert into examination(exam_id, name, diagnosis, age, gender) select exam_id + (select value from variable where name = 'max_exam_id'), name, diagnosis,age,gender from source.examination;

-- Paste measurements
insert into measurement(meas_id, time, exam_id) select meas_id + (select value from variable where name = 'max_meas_id'), time, exam_id + (select value from variable where name = 'max_exam_id') from source.measurement;

-- Paste signals
insert into signal(data, dt, meas_id, edited) select data, dt, meas_id + (select value from variable where name = 'max_meas_id'), edited from source.signal;

-- Connect SMEP and groups
insert into group_element(exam_id, group_id) select exam_id+(select value from variable where name = 'max_exam_id'), group_id + (select value from variable where name = 'max_group_id') from source.group_element;
"""



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Very simple implementation of SME model."""

class Signal:
    """Signal."""
    
    dt = None
    """Sample step."""
    x = []
    """Samples values."""
    
class Measurement:
    """Measurement.

    Measurement contains one or more signals. 
    
    """

    time = None
    """Measurement start time."""
    ss = []
    """Signals."""
    
class Examination:
    """Examination.

    Examination contains one or more measurements.

    """
    
    name = None
    """Name of patient."""
    
    age = None
    """Age of patient."""
    
    gender = None
    """Gender of patient."""
    
    diagnosis = None
    """Diagnosis."""
    
    ms = []
    """Measurements."""
    
def merge_exams(e1, e2):
    """Merge exams."""
    e = Examination()
    e = e1
    for m in e2.ms:
        e1.ms.append(m)
    return e
